@@ -4,27 +4,34 @@
 
 import os
 import re
+import sys
 import weakref
 from collections import defaultdict
+from threading import Thread
 from contextlib import contextmanager
+from functools import partial
 from html5_parser import parse
 from lxml import html
 from qt.core import (
     QAction, QApplication, QBrush, QByteArray, QCheckBox, QColor, QColorDialog, QDialog,
     QDialogButtonBox, QFont, QFontInfo, QFontMetrics, QFormLayout, QHBoxLayout, QIcon,
-    QKeySequence, QLabel, QLineEdit, QMenu, QPalette, QPlainTextEdit, QPushButton,
-    QSize, QSyntaxHighlighter, Qt, QTabWidget, QTextBlockFormat, QTextCharFormat,
-    QTextCursor, QTextEdit, QTextFormat, QTextListFormat, QTimer, QToolButton, QUrl,
+    QKeySequence, QLabel, QLineEdit, QMenu, QPalette, QPlainTextEdit, QPointF,
+    QPushButton, QSize, QSpinBox, QSyntaxHighlighter, Qt, QTabWidget, QTextBlockFormat,
+    QTextCharFormat, QTextCursor, QTextDocument, QTextEdit, QTextFormat,
+    QTextFrameFormat, QTextImageFormat, QTextListFormat, QTimer, QToolButton, QUrl,
     QVBoxLayout, QWidget, pyqtSignal, pyqtSlot,
 )
 
-from calibre import xml_replace_entities
+from calibre import browser, fit_image, xml_replace_entities
 from calibre.db.constants import DATA_DIR_NAME
 from calibre.ebooks.chardet import xml_to_unicode
 from calibre.gui2 import (
     NO_URL_FORMATTING, choose_dir, choose_files, error_dialog, gprefs, is_dark_theme,
+    question_dialog, safe_open_url,
 )
+from calibre.gui2 import FunctionDispatcher
 from calibre.gui2.book_details import resolved_css
+from calibre.gui2.dialogs.progress import ProgressDialog
 from calibre.gui2.flow_toolbar import create_flow_toolbar
 from calibre.gui2.widgets import LineEditECM
 from calibre.gui2.widgets2 import to_plain_text
@@ -36,6 +43,8 @@ from polyglot.builtins import iteritems, itervalues
 
 # Cleanup Qt markup {{{
 
+OBJECT_REPLACEMENT_CHAR = '\ufffc'
+
 
 def parse_style(style):
     props = filter(None, (x.strip() for x in style.split(';')))
@@ -209,6 +218,13 @@ def cleanup_qt_markup(root):
                 if ts:
                     remove_margins(li, ts)
                     remove_zero_indents(ts)
+    for img in root.xpath('//img[@style]'):
+        s = style_map.get(img)
+        if s:
+            if s == {'float': 'left'}:
+                s['margin-right'] = '0.5em'
+            elif s == {'float': 'right'}:
+                s['margin-left'] = '0.5em'
     for style in itervalues(style_map):
         filter_qt_styles(style)
         fw = style.get('font-weight')
@@ -226,7 +242,7 @@ def cleanup_qt_markup(root):
 # }}}
 
 
-def fix_html(original_html, original_txt, remove_comments=True):
+def fix_html(original_html, original_txt, remove_comments=True, callback=None):
     raw = original_html
     raw = xml_to_unicode(raw, strip_encoding_pats=True, resolve_entities=True)[0]
     if remove_comments:
@@ -248,6 +264,8 @@ def fix_html(original_html, original_txt, remove_comments=True):
     except Exception:
         import traceback
         traceback.print_exc()
+    if callback is not None:
+        callback(root, original_txt)
     elems = []
     for body in root.xpath('//body'):
         if body.text:
@@ -266,6 +284,8 @@ def fix_html(original_html, original_txt, remove_comments=True):
 class EditorWidget(QTextEdit, LineEditECM):  # {{{
 
     data_changed = pyqtSignal()
+    insert_images_separately = False
+    can_store_images = False
 
     @property
     def readonly(self):
@@ -341,8 +361,9 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
 
         r('color', 'format-text-color', _('Foreground color'))
         r('background', 'format-fill-color', _('Background color'))
-        r('insert_link', 'insert-link', _('Insert link or image'),
+        r('insert_link', 'insert-link', _('Insert link') if self.insert_images_separately else _('Insert link or image'),
           shortcut=QKeySequence('Ctrl+l', QKeySequence.SequenceFormat.PortableText))
+        r('insert_image', 'view-image', _('Insert image'), shortcut=QKeySequence('Ctrl+p', QKeySequence.SequenceFormat.PortableText))
         r('insert_hr', 'format-text-hr', _('Insert separator'),)
         r('clear', 'trash', _('Clear'))
 
@@ -544,9 +565,87 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
         self.focus_self()
 
     def do_paste(self):
+        images_before = set()
+        for fmt in self.document().allFormats():
+            if fmt.isImageFormat():
+                images_before.add(fmt.toImageFormat().name())
         self.paste()
+        if self.can_store_images:
+            added = set()
+            for fmt in self.document().allFormats():
+                if fmt.isImageFormat():
+                    name = fmt.toImageFormat().name()
+                    if name not in images_before and name.partition(':')[0] in ('https', 'http'):
+                        added.add(name)
+            if added and question_dialog(
+                self, _('Download images'), _(
+                    'Download all remote images in the pasted content?')):
+                self.download_images(added)
         self.focus_self()
 
+    def download_images(self, urls):
+        from calibre.web import get_download_filename_from_response
+        br = browser()
+        d = self.document()
+        c = self.textCursor()
+        c.setPosition(0)
+        pos_map = {}
+        while True:
+            c = d.find(OBJECT_REPLACEMENT_CHAR, c, QTextDocument.FindFlag.FindCaseSensitively)
+            if c.isNull():
+                break
+            fmt = c.charFormat()
+            if fmt.isImageFormat():
+                name = fmt.toImageFormat().name()
+                if name in urls:
+                    pos_map[name] = c.position()
+
+        pd = ProgressDialog(title=_('Downloading images...'), min=0, max=0, parent=self)
+        pd.canceled_signal.connect(pd.accept)
+        data_map = {}
+        error_map = {}
+        set_msg = FunctionDispatcher(pd.set_msg, parent=pd)
+        accept = FunctionDispatcher(pd.accept, parent=pd)
+
+        def do_download():
+            for url, pos in pos_map.items():
+                if pd.canceled:
+                    return
+                set_msg(_('Downloading {}...').format(url))
+                try:
+                    res = br.open_novisit(url, timeout=30)
+                    fname = get_download_filename_from_response(res)
+                    data = res.read()
+                except Exception as err:
+                    error_map[url] = err
+                else:
+                    data_map[url] = fname, data
+            accept()
+        Thread(target=do_download, daemon=True).start()
+        pd.exec()
+        if pd.canceled:
+            return
+
+        for url, pos in pos_map.items():
+            fname, data = data_map[url]
+            newname = self.commit_downloaded_image(data, fname)
+            c = self.textCursor()
+            c.setPosition(pos)
+            alignment = QTextFrameFormat.Position.InFlow
+            f = self.frame_for_cursor(c)
+            if f is not None:
+                alignment = f.frameFormat().position()
+            fmt = c.charFormat()
+            i = fmt.toImageFormat()
+            i.setName(newname)
+            c.deletePreviousChar()
+            c.insertImage(i, alignment)
+        if error_map:
+            m = '\n'.join(
+                _('Could not download {0} with error:').format(url) + '\n\t' + str(err) + '\n\n' for url, err in error_map.items())
+            error_dialog(self, _('Failed to download some images'), _(
+                'Some images could not be downloaded, click "Show details" to see which ones'), det_msg=m, show=True)
+
     def do_paste_and_match_style(self):
         text = QApplication.instance().clipboard().text()
         if text:
@@ -645,6 +744,14 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
             c.movePosition(QTextCursor.MoveOperation.EndOfBlock, QTextCursor.MoveMode.MoveAnchor)
             c.insertHtml('<hr>')
 
+    def do_insert_image(self):
+        from calibre.gui2 import choose_images
+        files = choose_images(self, 'choose-image-for-comments-editor', _('Choose image'), formats='png jpeg jpg gif svg webp'.split())
+        if files:
+            self.focus_self()
+            with self.editing_cursor() as c:
+                c.insertImage(files[0])
+
     def do_insert_link(self, *args):
         link, name, is_image = self.ask_link()
         if not link:
@@ -808,7 +915,10 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
 
     @property
     def html(self):
-        return fix_html(self.toHtml(), self.toPlainText().strip())
+        return fix_html(self.toHtml(), self.toPlainText().strip(), callback=self.get_html_callback)
+
+    def get_html_callback(self, root, text):
+        pass
 
     @html.setter
     def html(self, val):
@@ -823,20 +933,23 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
             if qurl.isRelative():
                 qurl = self.base_url.resolved(qurl)
             if qurl.isLocalFile():
+                data = None
                 path = qurl.toLocalFile()
                 try:
                     with open(path, 'rb') as f:
                         data = f.read()
                 except OSError:
                     if path.rpartition('.')[-1].lower() in {'jpg', 'jpeg', 'gif', 'png', 'bmp', 'webp'}:
-                        return QByteArray(bytearray.fromhex(
+                        data = bytearray.fromhex(
                                     '89504e470d0a1a0a0000000d49484452'
                                     '000000010000000108060000001f15c4'
                                     '890000000a49444154789c6300010000'
                                     '0500010d0a2db40000000049454e44ae'
-                                    '426082'))
-                else:
-                    return QByteArray(data)
+                                    '426082')
+                if data is not None:
+                    r = QByteArray(data)
+                    self.document().addResource(rtype, qurl, r)
+                    return r
 
     def set_html(self, val, allow_undo=True):
         if not allow_undo or self.readonly:
@@ -877,8 +990,138 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
         ans.setHtml(html)
         return ans
 
+    def remove_image_at(self, cursor_pos):
+        c = self.textCursor()
+        c.clearSelection()
+        c.setPosition(cursor_pos)
+        fmt = c.charFormat()
+        if fmt.isImageFormat():
+            c.deletePreviousChar()
+
+    def resize_image_at(self, cursor_pos):
+        c = self.textCursor()
+        c.clearSelection()
+        c.setPosition(cursor_pos)
+        c.movePosition(QTextCursor.MoveOperation.PreviousCharacter, QTextCursor.MoveMode.KeepAnchor)
+        fmt = c.charFormat()
+        if not fmt.isImageFormat():
+            return
+        fmt = fmt.toImageFormat()
+        from calibre.utils.img import image_from_data
+        img = image_from_data(self.loadResource(QTextDocument.ResourceType.ImageResource, QUrl(fmt.name())))
+        w, h = int(fmt.width()), int(fmt.height())
+        d = QDialog(self)
+        l = QVBoxLayout(d)
+        la = QLabel(_('Shrink image to fit within:'))
+        h = QHBoxLayout()
+        l.addLayout(h)
+        la = QLabel(_('&Width:'))
+        h.addWidget(la)
+        d.width = w = QSpinBox(self)
+        w.setRange(0, 10000), w.setSuffix(' px')
+        w.setValue(int(fmt.width()))
+        h.addWidget(w), la.setBuddy(w)
+        w.setSpecialValueText(' ')
+        la = QLabel(_('&Height:'))
+        h.addWidget(la)
+        d.height = w = QSpinBox(self)
+        w.setRange(0, 10000), w.setSuffix(' px')
+        w.setValue(int(fmt.height()))
+        h.addWidget(w), la.setBuddy(w)
+        w.setSpecialValueText(' ')
+        h.addStretch(10)
+        bb = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel)
+        bb.accepted.connect(d.accept), bb.rejected.connect(d.reject)
+        d.setWindowTitle(_('Enter new size for image'))
+        l.addWidget(bb)
+        d.resize(d.sizeHint())
+
+        if d.exec() == QDialog.DialogCode.Accepted:
+            page_width, page_height = (d.width.value() or sys.maxsize), (d.height.value() or sys.maxsize)
+            w, h = int(img.width()), int(img.height())
+            resized, nw, nh = fit_image(w, h, page_width, page_height)
+            if resized:
+                fmt.setWidth(nw), fmt.setHeight(nh)
+            else:
+                f = QTextImageFormat()
+                f.setName(fmt.name())
+                fmt = f
+            c.setCharFormat(fmt)
+
+    def align_image_at(self, cursor_pos, alignment):
+        c = self.textCursor()
+        c.clearSelection()
+        c.setPosition(cursor_pos)
+        fmt = c.charFormat()
+        if fmt.isImageFormat() and c.currentFrame():
+            f = self.frame_for_cursor(c)
+            if f is not None:
+                ff = f.frameFormat()
+                ff.setPosition(alignment)
+                f.setFrameFormat(ff)
+                self.document().markContentsDirty(cursor_pos-2, 5)
+            else:
+                c.deleteChar()
+                c.insertImage(fmt.toImageFormat(), alignment)
+
+    def first_image_replacement_char_position_for(self, image_name):
+        d = self.document()
+        c = self.textCursor()
+        c.setPosition(0)
+        while True:
+            c = d.find(OBJECT_REPLACEMENT_CHAR, c, QTextDocument.FindFlag.FindCaseSensitively)
+            if c.isNull():
+                break
+            fmt = c.charFormat()
+            if fmt.isImageFormat() and fmt.toImageFormat().name() == image_name:
+                return c.position()
+        return -1
+
+    def frame_for_cursor(self, c):
+        q = c.position()
+        for cf in c.currentFrame().childFrames():
+            a, b = cf.firstPosition(), cf.lastPosition()
+            a, b = min(a, b), max(a, b)
+            if a <= q <= b:
+                return cf
+
+    def open_link(self, link: str) -> None:
+        qurl = QUrl(link, QUrl.ParsingMode.TolerantMode)
+        if qurl.isRelative() and self.base_url:
+            qurl = QUrl.fromLocalFile(os.path.join(os.path.dirname(self.base_url.toLocalFile()), qurl.path()))
+        safe_open_url(qurl)
+
     def contextMenuEvent(self, ev):
         menu = QMenu(self)
+        img_name = self.document().documentLayout().imageAt(QPointF(ev.pos()))
+        if img_name:
+            pos = self.first_image_replacement_char_position_for(img_name)
+            if pos > -1:
+                c = self.textCursor()
+                c.clearSelection()
+                c.setPosition(pos)
+                pos = QTextFrameFormat.Position.InFlow
+                ff = self.frame_for_cursor(c)
+                if ff is not None:
+                    pos = ff.frameFormat().position()
+                align_menu = menu.addMenu(QIcon.ic('view-image.png'), _('Image...'))
+                def a(text, epos):
+                    ac = align_menu.addAction(text)
+                    ac.setCheckable(True)
+                    ac.triggered.connect(partial(self.align_image_at, c.position(), epos))
+                    if pos == epos:
+                        ac.setChecked(True)
+                cs = align_menu.addAction(QIcon.ic('resize.png'), _('Change size'))
+                cs.triggered.connect(partial(self.resize_image_at, c.position()))
+                align_menu.addSeparator()
+                a(_('Float to the left'), QTextFrameFormat.Position.FloatLeft)
+                a(_('Inline with text'), QTextFrameFormat.Position.InFlow)
+                a(_('Float to the right'), QTextFrameFormat.Position.FloatRight)
+                align_menu.addSeparator()
+                align_menu.addAction(QIcon.ic('trash.png'), _('Remove this image')).triggered.connect(partial(self.remove_image_at, c.position()))
+        link_name = self.document().documentLayout().anchorAt(QPointF(ev.pos()))
+        if link_name:
+            menu.addAction(QIcon.ic('insert-link.png'), _('Open link'), partial(self.open_link, link_name))
         for ac in 'undo redo -- cut copy paste paste_and_match_style -- select_all'.split():
             if ac == '--':
                 menu.addSeparator()
@@ -901,6 +1144,8 @@ class EditorWidget(QTextEdit, LineEditECM):  # {{{
         menu.addMenu(am)
         am.addAction(self.action_block_style)
         am.addAction(self.action_insert_link)
+        if self.insert_images_separately:
+            am.addAction(self.action_insert_image)
         am.addAction(self.action_background)
         am.addAction(self.action_color)
         menu.addAction(_('Smarten punctuation'), parent.smarten_punctuation)
@@ -1137,12 +1382,13 @@ class Editor(QWidget):  # {{{
 
     toolbar_prefs_name = None
     data_changed = pyqtSignal()
+    editor_class = EditorWidget
 
     def __init__(self, parent=None, one_line_toolbar=False, toolbar_prefs_name=None):
         QWidget.__init__(self, parent)
         self.toolbar_prefs_name = toolbar_prefs_name or self.toolbar_prefs_name
         self.toolbar = create_flow_toolbar(self, restrict_to_single_line=one_line_toolbar, icon_size=18)
-        self.editor = EditorWidget(self)
+        self.editor = self.editor_class(self)
         self.editor.data_changed.connect(self.data_changed)
         self.set_base_url = self.editor.set_base_url
         self.set_html = self.editor.set_html
@@ -1199,6 +1445,8 @@ class Editor(QWidget):  # {{{
 
         self.toolbar.add_action(self.editor.action_block_style, popup_mode=QToolButton.ToolButtonPopupMode.InstantPopup)
         self.toolbar.add_action(self.editor.action_insert_link)
+        if self.editor.insert_images_separately:
+            self.toolbar.add_action(self.editor.action_insert_image)
         self.toolbar.add_action(self.editor.action_insert_hr)
         self.toolbar.add_separator()
 
@@ -1304,5 +1552,7 @@ if __name__ == '__main__':
     set <u>out</u> to have an <em>affair</em>, <span style="font-style:italic; background-color:red">
     much</span> less a <s>long-term</s>, <b>devoted</b> one.</span><p>hello'''
     w.html = '<div><p id="moo" align="justify">Testing <em>a</em> link.</p><p align="justify">\xa0</p><p align="justify">ss</p></div>'
+    i = 'file:///home/kovid/work/calibre/resources/images/'
+    w.html = f'<p>Testing <img src="{i}/donate.png"> img and another <img src="{i}/lt.png">file</p>'
     app.exec()
     # print w.html
